use std::ffi::c_void;
use std::ptr;
use std::{fs::File, io::Read};

use serde::Deserialize;
use serde_with::base64::{Base64, Standard};
use serde_with::formats::Padded;
use serde_with::serde_as;

use windows_sys::Win32::Foundation::HMODULE;
use windows_sys::Win32::System::LibraryLoader;
use windows_sys::Win32::System::Memory;

#[serde_as]
#[derive(Debug, Deserialize)]
struct Import {
    #[serde(with = "serde_bytes")]
    dll: Vec<u8>,
    #[serde(with = "serde_bytes")]
    symbol: Vec<u8>,
    address: usize,
}

#[serde_as]
#[derive(Debug, Deserialize)]
struct Binary {
    #[serde(rename = "startingState")]
    #[serde_as(as = "Base64<Standard, Padded>")]
    starting_state: Vec<u8>,
    imports: Vec<Import>,
    #[serde(rename = "entryPoint")]
    entry_point: usize,
}
fn main() {
    unsafe { _main() }
}
unsafe fn _main() {
    let mut input = File::open("binaries/main.json").unwrap();
    let mut buf: Vec<u8> = vec![];
    let _ = input.read_to_end(&mut buf).unwrap();

    let binary = serde_json::from_slice::<Binary>(&buf).unwrap();
    println!("Binary: {:?}", binary);

    let address_space = Memory::VirtualAlloc(
        ptr::null(),
        binary.starting_state.len(),
        Memory::MEM_COMMIT | Memory::MEM_RESERVE,
        Memory::PAGE_EXECUTE_READWRITE,
    );
    address_space.copy_from(
        binary.starting_state.as_ptr() as *const c_void,
        binary.starting_state.len(),
    );
    for i in binary.imports.iter() {
        let mut lcpstr_dll = vec![];
        lcpstr_dll.extend(&i.dll);
        lcpstr_dll.push(0);

        let mut lpcstr_procname = vec![];
        lpcstr_procname.extend(&i.symbol);
        lpcstr_procname.push(0);

        println!("handling import");
        let dll: HMODULE = LibraryLoader::LoadLibraryA(lcpstr_dll.as_ptr());

        // TODO: Free anything? No real need to.
        let proc = LibraryLoader::GetProcAddress(dll, lpcstr_procname.as_ptr()).unwrap();
        let proc_ptr = proc as *mut c_void;
        println!(
            "writing at {} out of {}",
            i.address,
            binary.starting_state.len()
        );
        let location = address_space.byte_add(i.address) as *mut *mut c_void;
        *location = proc_ptr;
        println!("handled import");
    }
    println!("done with imports");
    let fn_start = address_space.byte_add(binary.entry_point);
    let function: unsafe fn() -> () = std::mem::transmute(fn_start);

    function();
    println!("returned");
}
